@@ -9,6 +9,7 @@ import importlib.util
 import inspect
 import itertools
 import os
+import os.path
 import platform
 import re
 import sys
@@ -17,15 +18,16 @@ import traceback
 
 from types import FrameType, ModuleType
 from typing import (
-    cast, Any, Iterable, List, Optional, Set, Tuple, Type, TYPE_CHECKING,
+    cast, Any, TYPE_CHECKING,
 )
+from collections.abc import Iterable
 
 from coverage import env
 from coverage.disposition import FileDisposition, disposition_init
 from coverage.exceptions import CoverageException, PluginError
 from coverage.files import TreeMatcher, GlobMatcher, ModuleMatcher
 from coverage.files import prep_patterns, find_python_files, canonical_filename
-from coverage.misc import sys_modules_saved
+from coverage.misc import isolate_module, sys_modules_saved
 from coverage.python import source_for_file, source_for_morf
 from coverage.types import TFileDisposition, TMorf, TWarnFn, TDebugCtl
 
@@ -38,7 +40,7 @@ if TYPE_CHECKING:
 # when deciding where the stdlib is.  These modules are not used for anything,
 # they are modules importable from the pypy lib directories, so that we can
 # find those directories.
-modules_we_happen_to_have: List[ModuleType] = [
+modules_we_happen_to_have: list[ModuleType] = [
     inspect, itertools, os, platform, re, sysconfig, traceback,
 ]
 
@@ -56,6 +58,8 @@ if env.PYPY:
         pass
 
 
+os = isolate_module(os)
+
 def canonical_path(morf: TMorf, directory: bool = False) -> str:
     """Return the canonical path of the module or file `morf`.
 
@@ -70,7 +74,7 @@ def canonical_path(morf: TMorf, directory: bool = False) -> str:
     return morf_path
 
 
-def name_for_module(filename: str, frame: Optional[FrameType]) -> str:
+def name_for_module(filename: str, frame: FrameType | None) -> str:
     """Get the name of the module for a filename and frame.
 
     For configurability's sake, we allow __main__ modules to be matched by
@@ -83,20 +87,16 @@ def name_for_module(filename: str, frame: Optional[FrameType]) -> str:
 
     """
     module_globals = frame.f_globals if frame is not None else {}
-    dunder_name: str = module_globals.get('__name__', None)
+    dunder_name: str = module_globals.get("__name__", None)
 
-    if isinstance(dunder_name, str) and dunder_name != '__main__':
+    if isinstance(dunder_name, str) and dunder_name != "__main__":
         # This is the usual case: an imported module.
         return dunder_name
 
-    loader = module_globals.get('__loader__', None)
-    for attrname in ('fullname', 'name'):   # attribute renamed in py3.2
-        if hasattr(loader, attrname):
-            fullname = getattr(loader, attrname)
-        else:
-            continue
-
-        if isinstance(fullname, str) and fullname != '__main__':
+    spec = module_globals.get("__spec__", None)
+    if spec:
+        fullname = spec.name
+        if isinstance(fullname, str) and fullname != "__main__":
             # Module loaded via: runpy -m
             return fullname
 
@@ -110,18 +110,18 @@ def name_for_module(filename: str, frame: Optional[FrameType]) -> str:
 
 def module_is_namespace(mod: ModuleType) -> bool:
     """Is the module object `mod` a PEP420 namespace module?"""
-    return hasattr(mod, '__path__') and getattr(mod, '__file__', None) is None
+    return hasattr(mod, "__path__") and getattr(mod, "__file__", None) is None
 
 
 def module_has_file(mod: ModuleType) -> bool:
     """Does the module object `mod` have an existing __file__ ?"""
-    mod__file__ = getattr(mod, '__file__', None)
+    mod__file__ = getattr(mod, "__file__", None)
     if mod__file__ is None:
         return False
     return os.path.exists(mod__file__)
 
 
-def file_and_path_for_module(modulename: str) -> Tuple[Optional[str], List[str]]:
+def file_and_path_for_module(modulename: str) -> tuple[str | None, list[str]]:
     """Find the file and search path for `modulename`.
 
     Returns:
@@ -142,7 +142,7 @@ def file_and_path_for_module(modulename: str) -> Tuple[Optional[str], List[str]]
     return filename, path
 
 
-def add_stdlib_paths(paths: Set[str]) -> None:
+def add_stdlib_paths(paths: set[str]) -> None:
     """Add paths where the stdlib can be found to the set `paths`."""
     # Look at where some standard modules are located. That's the
     # indication for "installed with the interpreter". In some
@@ -154,7 +154,7 @@ def add_stdlib_paths(paths: Set[str]) -> None:
             paths.add(canonical_path(m, directory=True))
 
 
-def add_third_party_paths(paths: Set[str]) -> None:
+def add_third_party_paths(paths: set[str]) -> None:
     """Add locations for third-party packages to the set `paths`."""
     # Get the paths that sysconfig knows about.
     scheme_names = set(sysconfig.get_scheme_names())
@@ -168,7 +168,7 @@ def add_third_party_paths(paths: Set[str]) -> None:
                 paths.add(config_paths[path_name])
 
 
-def add_coverage_paths(paths: Set[str]) -> None:
+def add_coverage_paths(paths: set[str]) -> None:
     """Add paths where coverage.py code can be found to the set `paths`."""
     cover_path = canonical_path(__file__, directory=True)
     paths.add(cover_path)
@@ -184,15 +184,15 @@ class InOrOut:
         self,
         config: CoverageConfig,
         warn: TWarnFn,
-        debug: Optional[TDebugCtl],
+        debug: TDebugCtl | None,
         include_namespace_packages: bool,
     ) -> None:
         self.warn = warn
         self.debug = debug
         self.include_namespace_packages = include_namespace_packages
 
-        self.source: List[str] = []
-        self.source_pkgs: List[str] = []
+        self.source: list[str] = []
+        self.source_pkgs: list[str] = []
         self.source_pkgs.extend(config.source_pkgs)
         for src in config.source or []:
             if os.path.isdir(src):
@@ -205,17 +205,17 @@ class InOrOut:
         self.omit = prep_patterns(config.run_omit)
 
         # The directories for files considered "installed with the interpreter".
-        self.pylib_paths: Set[str] = set()
+        self.pylib_paths: set[str] = set()
         if not config.cover_pylib:
             add_stdlib_paths(self.pylib_paths)
 
         # To avoid tracing the coverage.py code itself, we skip anything
         # located where we are.
-        self.cover_paths: Set[str] = set()
+        self.cover_paths: set[str] = set()
         add_coverage_paths(self.cover_paths)
 
         # Find where third-party packages are installed.
-        self.third_paths: Set[str] = set()
+        self.third_paths: set[str] = set()
         add_third_party_paths(self.third_paths)
 
         def _debug(msg: str) -> None:
@@ -262,7 +262,7 @@ class InOrOut:
         # Check if the source we want to measure has been installed as a
         # third-party package.
         # Is the source inside a third-party area?
-        self.source_in_third = False
+        self.source_in_third_paths = set()
         with sys_modules_saved():
             for pkg in self.source_pkgs:
                 try:
@@ -274,27 +274,28 @@ class InOrOut:
                 if modfile:
                     if self.third_match.match(modfile):
                         _debug(
-                            f"Source is in third-party because of source_pkg {pkg!r} at {modfile!r}"
+                            f"Source in third-party: source_pkg {pkg!r} at {modfile!r}",
                         )
-                        self.source_in_third = True
+                        self.source_in_third_paths.add(canonical_path(source_for_file(modfile)))
                 else:
                     for pathdir in path:
                         if self.third_match.match(pathdir):
                             _debug(
-                                f"Source is in third-party because of {pkg!r} path directory " +
-                                f"at {pathdir!r}"
+                                f"Source in third-party: {pkg!r} path directory at {pathdir!r}",
                             )
-                            self.source_in_third = True
+                            self.source_in_third_paths.add(pathdir)
 
         for src in self.source:
             if self.third_match.match(src):
-                _debug(f"Source is in third-party because of source directory {src!r}")
-                self.source_in_third = True
+                _debug(f"Source in third-party: source directory {src!r}")
+                self.source_in_third_paths.add(src)
+        self.source_in_third_match = TreeMatcher(self.source_in_third_paths, "source_in_third")
+        _debug(f"Source in third-party matching: {self.source_in_third_match}")
 
         self.plugins: Plugins
-        self.disp_class: Type[TFileDisposition] = FileDisposition
+        self.disp_class: type[TFileDisposition] = FileDisposition
 
-    def should_trace(self, filename: str, frame: Optional[FrameType] = None) -> TFileDisposition:
+    def should_trace(self, filename: str, frame: FrameType | None = None) -> TFileDisposition:
         """Decide whether to trace execution in `filename`, with a reason.
 
         This function is called from the trace function.  As each new file name
@@ -312,8 +313,8 @@ class InOrOut:
             disp.reason = reason
             return disp
 
-        if original_filename.startswith('<'):
-            return nope(disp, "not a real original file name")
+        if original_filename.startswith("<"):
+            return nope(disp, "original file name is not real")
 
         if frame is not None:
             # Compiled Python files have two file names: frame.f_code.co_filename is
@@ -322,10 +323,11 @@ class InOrOut:
             # .pyc files can be moved after compilation (for example, by being
             # installed), we look for __file__ in the frame and prefer it to the
             # co_filename value.
-            dunder_file = frame.f_globals and frame.f_globals.get('__file__')
+            dunder_file = frame.f_globals and frame.f_globals.get("__file__")
             if dunder_file:
-                filename = source_for_file(dunder_file)
-                if original_filename and not original_filename.startswith('<'):
+                # Danger: __file__ can (rarely?) be of type Path.
+                filename = source_for_file(str(dunder_file))
+                if original_filename and not original_filename.startswith("<"):
                     orig = os.path.basename(original_filename)
                     if orig != os.path.basename(filename):
                         # Files shouldn't be renamed when moved. This happens when
@@ -337,15 +339,15 @@ class InOrOut:
             # Empty string is pretty useless.
             return nope(disp, "empty string isn't a file name")
 
-        if filename.startswith('memory:'):
+        if filename.startswith("memory:"):
             return nope(disp, "memory isn't traceable")
 
-        if filename.startswith('<'):
+        if filename.startswith("<"):
             # Lots of non-file execution is represented with artificial
             # file names like "<string>", "<doctest readme.txt[0]>", or
             # "<exec_function>".  Don't ever trace these executions, since we
             # can't do anything with the data later anyway.
-            return nope(disp, "not a real file name")
+            return nope(disp, "file name is not real")
 
         canonical = canonical_filename(filename)
         disp.canonical_filename = canonical
@@ -366,7 +368,7 @@ class InOrOut:
                         disp.has_dynamic_filename = True
                     else:
                         disp.source_filename = canonical_filename(
-                            file_tracer.source_filename()
+                            file_tracer.source_filename(),
                         )
                     break
             except Exception:
@@ -383,7 +385,7 @@ class InOrOut:
         if not disp.has_dynamic_filename:
             if not disp.source_filename:
                 raise PluginError(
-                    f"Plugin {plugin!r} didn't set source_filename for '{disp.original_filename}'"
+                    f"Plugin {plugin!r} didn't set source_filename for '{disp.original_filename}'",
                 )
             reason = self.check_include_omit_etc(disp.source_filename, frame)
             if reason:
@@ -391,7 +393,7 @@ class InOrOut:
 
         return disp
 
-    def check_include_omit_etc(self, filename: str, frame: Optional[FrameType]) -> Optional[str]:
+    def check_include_omit_etc(self, filename: str, frame: FrameType | None) -> str | None:
         """Check a file name against the include, omit, etc, rules.
 
         Returns a string or None.  String means, don't trace, and is the reason
@@ -419,9 +421,8 @@ class InOrOut:
                     ok = True
             if not ok:
                 return extra + "falls outside the --source spec"
-            if not self.source_in_third:
-                if self.third_match.match(filename):
-                    return "inside --source, but is third-party"
+            if self.third_match.match(filename) and not self.source_in_third_match.match(filename):
+                return "inside --source, but is third-party"
         elif self.include_match:
             if not self.include_match.match(filename):
                 return "falls outside the --include trees"
@@ -484,11 +485,11 @@ class InOrOut:
                     msg = f"Already imported a file that will be measured: {filename}"
                     self.warn(msg, slug="already-imported")
                     warned.add(filename)
-                elif self.debug and self.debug.should('trace'):
+                elif self.debug and self.debug.should("trace"):
                     self.debug.write(
                         "Didn't trace already imported file {!r}: {}".format(
-                            disp.original_filename, disp.reason
-                        )
+                            disp.original_filename, disp.reason,
+                        ),
                     )
 
     def warn_unimported_source(self) -> None:
@@ -522,13 +523,13 @@ class InOrOut:
         msg = f"Module {pkg} was previously imported, but not measured"
         self.warn(msg, slug="module-not-measured")
 
-    def find_possibly_unexecuted_files(self) -> Iterable[Tuple[str, Optional[str]]]:
+    def find_possibly_unexecuted_files(self) -> Iterable[tuple[str, str | None]]:
         """Find files in the areas of interest that might be untraced.
 
         Yields pairs: file path, and responsible plug-in name.
         """
         for pkg in self.source_pkgs:
-            if (not pkg in sys.modules or
+            if (pkg not in sys.modules or
                 not module_has_file(sys.modules[pkg])):
                 continue
             pkg_file = source_for_file(cast(str, sys.modules[pkg].__file__))
@@ -537,13 +538,13 @@ class InOrOut:
         for src in self.source:
             yield from self._find_executable_files(src)
 
-    def _find_plugin_files(self, src_dir: str) -> Iterable[Tuple[str, str]]:
+    def _find_plugin_files(self, src_dir: str) -> Iterable[tuple[str, str]]:
         """Get executable files from the plugins."""
         for plugin in self.plugins.file_tracers:
             for x_file in plugin.find_executable_files(src_dir):
                 yield x_file, plugin._coverage_plugin_name
 
-    def _find_executable_files(self, src_dir: str) -> Iterable[Tuple[str, Optional[str]]]:
+    def _find_executable_files(self, src_dir: str) -> Iterable[tuple[str, str | None]]:
         """Find executable files in `src_dir`.
 
         Search for files in `src_dir` that can be executed because they
@@ -567,7 +568,7 @@ class InOrOut:
                 continue
             yield file_path, plugin_name
 
-    def sys_info(self) -> Iterable[Tuple[str, Any]]:
+    def sys_info(self) -> Iterable[tuple[str, Any]]:
         """Our information for Coverage.sys_info.
 
         Returns a list of (key, value) pairs.
@@ -576,12 +577,13 @@ class InOrOut:
             ("coverage_paths", self.cover_paths),
             ("stdlib_paths", self.pylib_paths),
             ("third_party_paths", self.third_paths),
+            ("source_in_third_party_paths", self.source_in_third_paths),
         ]
 
         matcher_names = [
-            'source_match', 'source_pkgs_match',
-            'include_match', 'omit_match',
-            'cover_match', 'pylib_match', 'third_match',
+            "source_match", "source_pkgs_match",
+            "include_match", "omit_match",
+            "cover_match", "pylib_match", "third_match", "source_in_third_match",
         ]
 
         for matcher_name in matcher_names:
@@ -589,7 +591,7 @@ class InOrOut:
             if matcher:
                 matcher_info = matcher.info()
             else:
-                matcher_info = '-none-'
+                matcher_info = "-none-"
             info.append((matcher_name, matcher_info))
 
         return info
